# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot.interfaces import IRenderable
from buildbot.interfaces import ITriggerableScheduler
from buildbot.process.buildstep import CANCELLED
from buildbot.process.buildstep import EXCEPTION
from buildbot.process.buildstep import SUCCESS
from buildbot.process.buildstep import BuildStep
from buildbot.process.properties import Properties
from buildbot.process.properties import Property
from buildbot.process.results import ALL_RESULTS
from buildbot.process.results import statusToString
from buildbot.process.results import worst_status
from buildbot.reporters.utils import getURLForBuild
from buildbot.reporters.utils import getURLForBuildrequest


class Trigger(BuildStep):
    name = "trigger"

    renderables = [
        'alwaysUseLatest',
        'parent_relationship',
        'schedulerNames',
        'set_properties',
        'sourceStamps',
        'updateSourceStamp',
        'waitForFinish',
    ]

    flunkOnFailure = True

    def __init__(
        self,
        schedulerNames=None,
        sourceStamp=None,
        sourceStamps=None,
        updateSourceStamp=None,
        alwaysUseLatest=False,
        waitForFinish=False,
        set_properties=None,
        copy_properties=None,
        parent_relationship="Triggered from",
        unimportantSchedulerNames=None,
        **kwargs,
    ):
        if schedulerNames is None:
            schedulerNames = []
        if unimportantSchedulerNames is None:
            unimportantSchedulerNames = []
        if not schedulerNames:
            config.error("You must specify a scheduler to trigger")
        if (sourceStamp or sourceStamps) and (updateSourceStamp is not None):
            config.error("You can't specify both sourceStamps and updateSourceStamp")
        if (sourceStamp or sourceStamps) and alwaysUseLatest:
            config.error("You can't specify both sourceStamps and alwaysUseLatest")
        if alwaysUseLatest and (updateSourceStamp is not None):
            config.error("You can't specify both alwaysUseLatest and updateSourceStamp")

        def hasRenderable(l):
            for s in l:
                if IRenderable.providedBy(s):
                    return True
            return False

        if not hasRenderable(schedulerNames) and not hasRenderable(unimportantSchedulerNames):
            if not set(schedulerNames).issuperset(set(unimportantSchedulerNames)):
                config.error("unimportantSchedulerNames must be a subset of schedulerNames")

        self.schedulerNames = schedulerNames
        self.unimportantSchedulerNames = unimportantSchedulerNames
        self.sourceStamps = sourceStamps or []
        if sourceStamp:
            self.sourceStamps.append(sourceStamp)
        if updateSourceStamp is not None:
            self.updateSourceStamp = updateSourceStamp
        else:
            self.updateSourceStamp = not (alwaysUseLatest or self.sourceStamps)
        self.alwaysUseLatest = alwaysUseLatest
        self.waitForFinish = waitForFinish

        if set_properties is None:
            set_properties = {}
        if copy_properties is None:
            copy_properties = []

        properties = {}
        properties.update(set_properties)
        for i in copy_properties:
            properties[i] = Property(i)
        self.set_properties = properties
        self.parent_relationship = parent_relationship
        self.running = False
        self.ended = False
        self.brids = []
        self.triggeredNames = None
        self.waitForFinishDeferred = None
        self._result_list = []
        super().__init__(**kwargs)

    def interrupt(self, reason):
        # We cancel the buildrequests, as the data api handles
        # both cases:
        # - build started: stop is sent,
        # - build not created yet: related buildrequests are set to CANCELLED.
        # Note that there is an identified race condition though (more details
        # are available at buildbot.data.buildrequests).
        for brid in self.brids:
            self.master.data.control(
                "cancel", {'reason': 'parent build was interrupted'}, ("buildrequests", brid)
            )
        if self.running and not self.ended:
            self.ended = True
            # if we are interrupted because of a connection lost, we interrupt synchronously
            if self.build.conn is None and self.waitForFinishDeferred is not None:
                self.waitForFinishDeferred.cancel()

    # Create the properties that are used for the trigger
    def createTriggerProperties(self, properties):
        # make a new properties object from a dict rendered by the old
        # properties object
        trigger_properties = Properties()
        trigger_properties.update(properties, "Trigger")
        return trigger_properties

    def getSchedulerByName(self, name):
        # we use the fact that scheduler_manager is a multiservice, with schedulers as childs
        # this allow to quickly find schedulers instance by name
        schedulers = self.master.scheduler_manager.namedServices
        if name not in schedulers:
            raise ValueError(f"unknown triggered scheduler: {name!r}")
        sch = schedulers[name]
        if not ITriggerableScheduler.providedBy(sch):
            raise ValueError(f"triggered scheduler is not ITriggerableScheduler: {name!r}")
        return sch

    # This customization endpoint allows users to dynamically select which
    # scheduler and properties to trigger
    def getSchedulersAndProperties(self):
        return [
            {
                'sched_name': sched,
                'props_to_set': self.set_properties,
                'unimportant': sched in self.unimportantSchedulerNames,
            }
            for sched in self.schedulerNames
        ]

    def prepareSourcestampListForTrigger(self):
        if self.sourceStamps:
            ss_for_trigger = {}
            for ss in self.sourceStamps:
                codebase = ss.get('codebase', '')
                assert codebase not in ss_for_trigger, "codebase specified multiple times"
                ss_for_trigger[codebase] = ss
            trigger_values = [ss_for_trigger[k] for k in sorted(ss_for_trigger.keys())]
            return trigger_values

        if self.alwaysUseLatest:
            return []

        # start with the sourcestamps from current build
        ss_for_trigger = {}
        objs_from_build = self.build.getAllSourceStamps()
        for ss in objs_from_build:
            ss_for_trigger[ss.codebase] = ss.asDict()

        # overrule revision in sourcestamps with got revision
        if self.updateSourceStamp:
            got = self.getAllGotRevisions()
            for codebase, ss in ss_for_trigger.items():
                if codebase in got:
                    ss['revision'] = got[codebase]

        trigger_values = [ss_for_trigger[k] for k in sorted(ss_for_trigger.keys())]
        return trigger_values

    def getAllGotRevisions(self):
        all_got_revisions = self.getProperty('got_revision', {})
        # For backwards compatibility all_got_revisions is a string if codebases
        # are not used. Convert to the default internal type (dict)
        if not isinstance(all_got_revisions, dict):
            all_got_revisions = {'': all_got_revisions}
        return all_got_revisions

    @defer.inlineCallbacks
    def worstStatus(self, overall_results, rclist, unimportant_brids):
        for was_cb, results in rclist:
            if isinstance(results, tuple):
                results, brids_dict = results

                # brids_dict.values() represents the list of brids kicked by a certain scheduler.
                # We want to ignore the result of ANY brid that was kicked off
                # by an UNimportant scheduler.
                if set(unimportant_brids).issuperset(set(brids_dict.values())):
                    continue

            if not was_cb:
                yield self.addLogWithFailure(results)
                results = EXCEPTION

            overall_results = worst_status(overall_results, results)
        return overall_results

    @defer.inlineCallbacks
    def addBuildUrls(self, rclist):
        brids = {}
        for was_cb, results in rclist:
            if isinstance(results, tuple):
                results, brids = results
            builderNames = {}
            if was_cb:  # errors were already logged in worstStatus
                for builderid, br in brids.items():
                    builds = yield self.master.db.builds.getBuilds(buildrequestid=br)
                    for build in builds:
                        builderid = build.builderid
                        # When virtual builders are used, the builderid used for triggering
                        # is not the same as the one that the build actually got
                        if builderid not in builderNames:
                            builderDict = yield self.master.data.get(("builders", builderid))
                            builderNames[builderid] = builderDict["name"]
                        num = build.number
                        url = getURLForBuild(self.master, builderid, num)
                        yield self.addURL(
                            f'{statusToString(build.results)}: {builderNames[builderid]} #{num}',
                            url,
                        )

    @defer.inlineCallbacks
    def _add_results(self, brid):
        @defer.inlineCallbacks
        def _is_buildrequest_complete(brid):
            buildrequest = yield self.master.db.buildrequests.getBuildRequest(brid)
            return buildrequest.complete

        event = ('buildrequests', str(brid), 'complete')
        yield self.master.mq.waitUntilEvent(event, lambda: _is_buildrequest_complete(brid))
        builds = yield self.master.db.builds.getBuilds(buildrequestid=brid)
        for build in builds:
            self._result_list.append(build.results)
        self.updateSummary()

    @defer.inlineCallbacks
    def run(self):
        schedulers_and_props = yield self.getSchedulersAndProperties()

        schedulers_and_props_list = []

        # To be back compatible we need to differ between old and new style
        # schedulers_and_props can either consist of 2 elements tuple or
        # dictionary
        for element in schedulers_and_props:
            if isinstance(element, dict):
                schedulers_and_props_list = schedulers_and_props
                break
            # Old-style back compatibility: Convert tuple to dict and make
            # it important
            d = {'sched_name': element[0], 'props_to_set': element[1], 'unimportant': False}
            schedulers_and_props_list.append(d)

        # post process the schedulernames, and raw properties
        # we do this out of the loop, as this can result in errors
        schedulers_and_props = [
            (
                self.getSchedulerByName(entry_dict['sched_name']),
                self.createTriggerProperties(entry_dict['props_to_set']),
                entry_dict['unimportant'],
            )
            for entry_dict in schedulers_and_props_list
        ]

        ss_for_trigger = self.prepareSourcestampListForTrigger()

        dl = []
        triggeredNames = []
        results = SUCCESS
        self.running = True

        unimportant_brids = []

        for sch, props_to_set, unimportant in schedulers_and_props:
            idsDeferred, resultsDeferred = sch.trigger(
                waited_for=self.waitForFinish,
                sourcestamps=ss_for_trigger,
                set_props=props_to_set,
                parent_buildid=self.build.buildid,
                parent_relationship=self.parent_relationship,
            )
            # we are not in a hurry of starting all in parallel and managing
            # the deferred lists, just let the db writes be serial.
            brids = {}
            try:
                _, brids = yield idsDeferred
            except Exception as e:
                yield self.addLogWithException(e)
                results = EXCEPTION
            if unimportant:
                unimportant_brids.extend(brids.values())
            self.brids.extend(brids.values())
            for brid in brids.values():
                # put the url to the brids, so that we can have the status from
                # the beginning
                url = getURLForBuildrequest(self.master, brid)
                yield self.addURL(f"{sch.name} #{brid}", url)
                # No yield since we let this happen as the builds complete
                self._add_results(brid)

            dl.append(resultsDeferred)
            triggeredNames.append(sch.name)
            if self.ended:
                return CANCELLED
        self.triggeredNames = triggeredNames

        if self.waitForFinish:
            self.waitForFinishDeferred = defer.DeferredList(dl, consumeErrors=True)
            try:
                rclist = yield self.waitForFinishDeferred
            except defer.CancelledError:
                pass
            # we were interrupted, don't bother update status
            if self.ended:
                return CANCELLED
            yield self.addBuildUrls(rclist)
            results = yield self.worstStatus(results, rclist, unimportant_brids)
        else:
            # do something to handle errors
            for d in dl:
                d.addErrback(log.err, '(ignored) while invoking Triggerable schedulers:')

        return results

    def getResultSummary(self):
        if self.ended:
            return {'step': 'interrupted'}
        return {'step': self.getCurrentSummary()['step']} if self.triggeredNames else {}

    def getCurrentSummary(self):
        if not self.triggeredNames:
            return {'step': 'running'}
        summary = ""
        if self._result_list:
            for status in ALL_RESULTS:
                count = self._result_list.count(status)
                if count:
                    summary = summary + (
                        f", {self._result_list.count(status)} {statusToString(status, count)}"
                    )
        return {'step': f"triggered {', '.join(self.triggeredNames)}{summary}"}



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Pattern
from typing import Sequence

from buildbot.util import ComparableMixin
from buildbot.util import NotABranch
from buildbot.util.ssfilter import _create_branch_filters
from buildbot.util.ssfilter import _create_filters
from buildbot.util.ssfilter import _create_property_filters


class ChangeFilter(ComparableMixin):
    # NOTE: If users use a filter_fn, we have no way to determine whether it has
    # changed at reconfig, so the scheduler will always be restarted.  That's as
    # good as Python can do.
    compare_attrs: ClassVar[Sequence[str]] = (
        'filter_fn',
        'filters',
        'property_filters',
        'project_fn',
        'repository_fn',
        'branch_fn',
        'category_fn',
        'codebase_fn',
    )

    def __init__(
        self,
        filter_fn: Callable[[Any], bool] | None = None,  # gets a Change object, returns boolean
        project: str | list[str] | None = None,
        project_not_eq: str | list[str] | None = None,
        project_re: str | Pattern | None = None,
        project_not_re: str | Pattern | None = None,
        project_fn: Callable[[str], bool] | None = None,
        repository: str | list[str] | None = None,
        repository_not_eq: str | list[str] | None = None,
        repository_re: str | Pattern | None = None,
        repository_not_re: str | Pattern | None = None,
        repository_fn: Callable[[str], bool] | None = None,
        branch: Any = NotABranch,
        branch_not_eq: Any = NotABranch,
        branch_re: str | Pattern | None = None,
        branch_not_re: str | Pattern | None = None,
        branch_fn: Callable[[str], bool] | None = None,
        category: str | list[str] | None = None,
        category_not_eq: str | list[str] | None = None,
        category_re: str | Pattern | None = None,
        category_not_re: str | Pattern | None = None,
        category_fn: Callable[[str], bool] | None = None,
        codebase: str | list[str] | None = None,
        codebase_not_eq: str | list[str] | None = None,
        codebase_re: str | Pattern | None = None,
        codebase_not_re: str | Pattern | None = None,
        codebase_fn: Callable[[str], bool] | None = None,
        property_eq: dict[str, Any] | None = None,
        property_not_eq: dict[str, Any] | None = None,
        property_re: dict[str, str | Pattern] | None = None,
        property_not_re: dict[str, str | Pattern] | None = None,
    ) -> None:
        self.filter_fn = filter_fn
        self.project_fn = project_fn
        self.repository_fn = repository_fn
        self.branch_fn = branch_fn
        self.category_fn = category_fn
        self.codebase_fn = codebase_fn

        self.filters = _create_filters(
            project, project_not_eq, project_re, project_not_re, 'project'
        )
        self.filters += _create_filters(
            repository,
            repository_not_eq,
            repository_re,
            repository_not_re,
            'repository',
        )
        self.filters += _create_branch_filters(
            branch,
            branch_not_eq,
            branch_re,
            branch_not_re,
            'branch',
        )
        self.filters += _create_filters(
            category,
            category_not_eq,
            category_re,
            category_not_re,
            'category',
        )
        self.filters += _create_filters(
            codebase,
            codebase_not_eq,
            codebase_re,
            codebase_not_re,
            'codebase',
        )
        self.property_filters = _create_property_filters(
            property_eq, property_not_eq, property_re, property_not_re, 'property'
        )

    def filter_change(self, change: Any) -> bool:
        if self.filter_fn is not None and not self.filter_fn(change):
            return False
        if self.project_fn is not None and not self.project_fn(change.project):
            return False
        if self.codebase_fn is not None and not self.codebase_fn(change.codebase):
            return False
        if self.repository_fn is not None and not self.repository_fn(change.repository):
            return False
        if self.category_fn is not None and not self.category_fn(change.category):
            return False
        if self.branch_fn is not None and not self.branch_fn(change.branch):
            return False

        for filter in self.filters:
            value = getattr(change, filter.prop, '')
            if not filter.is_matched(value):
                return False
        for filter in self.property_filters:
            value = change.properties.getProperty(filter.prop, '')
            if not filter.is_matched(value):
                return False
        return True

    def _get_repr_filters(self) -> list[str]:
        filters = []
        if self.filter_fn is not None:
            filters.append(f'{self.filter_fn.__name__}()')
        if self.project_fn is not None:
            filters.append(f'{self.project_fn.__name__}(project)')
        if self.codebase_fn is not None:
            filters.append(f'{self.codebase_fn.__name__}(codebase)')
        if self.repository_fn is not None:
            filters.append(f'{self.repository_fn.__name__}(repository)')
        if self.category_fn is not None:
            filters.append(f'{self.category_fn.__name__}(category)')
        if self.branch_fn is not None:
            filters.append(f'{self.branch_fn.__name__}(branch)')
        filters += [filter.describe() for filter in self.filters]
        filters += [filter.describe() for filter in self.property_filters]
        return filters

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} on {' and '.join(self._get_repr_filters())}>"

    @staticmethod
    def fromSchedulerConstructorArgs(
        change_filter: ChangeFilter | None = None,
        branch: Any = NotABranch,
        categories: str | list[str] | None = None,
    ) -> ChangeFilter | None:
        """
        Static method to create a filter based on constructor args
        change_filter, branch, and categories; use default values @code{None},
        @code{NotABranch}, and @code{None}, respectively.  These arguments are
        interpreted as documented for the
        L{buildbot.schedulers.basic.Scheduler} class.

        @returns: L{ChangeFilter} instance or None for not filtering
        """

        # use a change_filter, if given one
        if change_filter:
            if branch is not NotABranch or categories is not None:
                raise RuntimeError("cannot specify both change_filter and branch or categories")
            return change_filter
        elif branch is not NotABranch or categories:
            # build a change filter from the deprecated category and branch
            # args
            cfargs = {}
            if branch is not NotABranch:
                cfargs['branch'] = branch
            if categories:
                cfargs['category'] = categories
            return ChangeFilter(**cfargs)
        else:
            return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import inspect
import time
import traceback
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer
from twisted.internet import threads
from twisted.python import log

from buildbot import util
from buildbot.db.buildrequests import AlreadyClaimedError
from buildbot.db.buildsets import AlreadyCompleteError
from buildbot.db.changesources import ChangeSourceAlreadyClaimedError
from buildbot.db.logs import LogSlugExistsError
from buildbot.db.schedulers import SchedulerAlreadyClaimedError
from buildbot.process import metrics
from buildbot.util.sautils import get_sqlite_version

if TYPE_CHECKING:
    from typing import Callable
    from typing import TypeVar

    from typing_extensions import Concatenate
    from typing_extensions import ParamSpec

    _T = TypeVar('_T')
    _P = ParamSpec('_P')

# set this to True for *very* verbose query debugging output; this can
# be monkey-patched from master.cfg, too:
#     from buildbot.db import pool
#     pool.debug = True
debug = False
_debug_id = 1


def timed_do_fn(f):
    """Decorate a do function to log before, after, and elapsed time,
    with the name of the calling function.  This is not speedy!"""

    def wrap(callable, *args, **kwargs):
        global _debug_id

        # get a description of the function that called us
        st = traceback.extract_stack(limit=2)
        file, line, name, _ = st[0]

        # and its locals
        frame = inspect.currentframe()
        locals = frame.f_locals

        # invent a unique ID for the description
        id = _debug_id
        _debug_id = _debug_id + 1

        descr = f"{name}-{id:08x}"

        start_time = time.time()
        log.msg(f"{descr} - before ('{file}' line {line})")
        for name in locals:
            if name in ('self', 'thd'):
                continue
            log.msg(f"{descr} - {name} = {locals[name]!r}")

        # wrap the callable to log the begin and end of the actual thread
        # function
        def callable_wrap(*args, **kargs):
            log.msg(f"{descr} - thd start")
            try:
                return callable(*args, **kwargs)
            finally:
                log.msg(f"{descr} - thd end")

        d = f(callable_wrap, *args, **kwargs)

        @d.addBoth
        def after(x):
            end_time = time.time()
            elapsed = (end_time - start_time) * 1000
            log.msg(f"{descr} - after ({elapsed:0.2f} ms elapsed)")
            return x

        return d

    wrap.__name__ = f.__name__
    wrap.__doc__ = f.__doc__
    return wrap


class DBThreadPool:
    running = False

    def __init__(self, engine, reactor, verbose=False):
        # verbose is used by upgrade scripts, and if it is set we should print
        # messages about versions and other warnings
        log_msg = log.msg
        if verbose:

            def _log_msg(m):
                print(m)

            log_msg = _log_msg

        self.reactor = reactor

        pool_size = 5

        # If the engine has an C{optimal_thread_pool_size} attribute, then the
        # maxthreads of the thread pool will be set to that value.  This is
        # most useful for SQLite in-memory connections, where exactly one
        # connection (and thus thread) should be used.
        if hasattr(engine, 'optimal_thread_pool_size'):
            pool_size = engine.optimal_thread_pool_size

        self._pool = util.twisted.ThreadPool(
            minthreads=1, maxthreads=pool_size, name='DBThreadPool'
        )

        self.engine = engine
        if engine.dialect.name == 'sqlite':
            vers = get_sqlite_version()
            if vers < (3, 7):
                log_msg(f"Using SQLite Version {vers}")
                log_msg(
                    "NOTE: this old version of SQLite does not support "
                    "WAL journal mode; a busy master may encounter "
                    "'Database is locked' errors.  Consider upgrading."
                )
                if vers < (3, 6, 19):
                    log_msg("NOTE: this old version of SQLite is not supported.")
                    raise RuntimeError("unsupported SQLite version")

        # patch the do methods to do verbose logging if necessary
        if debug:
            self.do = timed_do_fn(self.do)
            self.do_with_engine = timed_do_fn(self.do_with_engine)

        self.forbidded_callable_return_type = self.get_sqlalchemy_result_type()

    def get_sqlalchemy_result_type(self):
        try:
            from sqlalchemy.engine import ResultProxy  # sqlalchemy 1.x - 1.3

            return ResultProxy
        except ImportError:
            pass

        try:
            from sqlalchemy.engine import Result  # sqlalchemy 1.4 and newer

            return Result
        except ImportError:
            pass

        raise ImportError("Could not import SQLAlchemy result type")

    def start(self):
        if not self.running:
            self._pool.start()
            self.running = True

    @defer.inlineCallbacks
    def stop(self):
        if self.running:
            yield threads.deferToThreadPool(self.reactor, self._pool, self.engine.dispose)
            self._pool.stop()
            self.running = False

    # Try about 170 times over the space of a day, with the last few tries
    # being about an hour apart.  This is designed to span a reasonable amount
    # of time for repairing a broken database server, while still failing
    # actual problematic queries eventually
    BACKOFF_START = 1.0
    BACKOFF_MULT = 1.05
    MAX_OPERATIONALERROR_TIME = 3600 * 24  # one day

    def __thd(
        self,
        with_engine: bool,
        callable: Callable[Concatenate[sa.engine.Engine | sa.engine.Connection, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T:
        # try to call callable(arg, *args, **kwargs) repeatedly until no
        # OperationalErrors occur, where arg is either the engine (with_engine)
        # or a connection (not with_engine)
        backoff = self.BACKOFF_START
        start = time.time()
        while True:
            if with_engine:
                arg = self.engine
            else:
                arg = self.engine.connect()
            try:
                try:
                    rv = callable(arg, *args, **kwargs)
                    assert not isinstance(rv, self.forbidded_callable_return_type), (
                        "do not return ResultProxy objects!"
                    )
                except sa.exc.OperationalError as e:
                    if not self.engine.should_retry(e):
                        log.err(e, 'Got fatal OperationalError on DB')
                        raise
                    elapsed = time.time() - start
                    if elapsed > self.MAX_OPERATIONALERROR_TIME:
                        log.err(
                            e,
                            f'Raising due to {self.MAX_OPERATIONALERROR_TIME} '
                            'seconds delay on DB query retries',
                        )
                        raise

                    metrics.MetricCountEvent.log("DBThreadPool.retry-on-OperationalError")
                    # sleep (remember, we're in a thread..)
                    time.sleep(backoff)
                    backoff *= self.BACKOFF_MULT
                    # and re-try
                    log.err(e, f'retrying {callable} after sql error {e}')
                    continue
                except Exception as e:
                    # AlreadyClaimedError are normal especially in a multimaster
                    # configuration
                    if not isinstance(
                        e,
                        (
                            AlreadyClaimedError,
                            ChangeSourceAlreadyClaimedError,
                            SchedulerAlreadyClaimedError,
                            AlreadyCompleteError,
                            LogSlugExistsError,
                        ),
                    ):
                        log.err(e, 'Got fatal Exception on DB')
                    raise
            finally:
                if not with_engine:
                    arg.close()
            break
        return rv

    def do_with_transaction(
        self,
        callable: Callable[Concatenate[sa.engine.Connection, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> defer.Deferred[_T]:
        """Same as `do`, but will wrap callable with `with conn.begin():`"""

        def _transaction(
            conn: sa.engine.Connection,
            callable: Callable[Concatenate[sa.engine.Connection, _P], _T],
            *args: _P.args,
            **kwargs: _P.kwargs,
        ) -> _T:
            with conn.begin():
                return callable(conn, *args, **kwargs)

        return self.do(_transaction, callable, *args, **kwargs)

    def do(
        self,
        callable: Callable[Concatenate[sa.engine.Connection, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> defer.Deferred[_T]:
        return threads.deferToThreadPool(
            self.reactor,
            self._pool,
            self.__thd,  # type: ignore[arg-type]
            False,
            callable,
            *args,
            **kwargs,
        )

    def do_with_engine(
        self,
        callable: Callable[Concatenate[sa.engine.Engine, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> defer.Deferred[_T]:
        return threads.deferToThreadPool(
            self.reactor,
            self._pool,
            self.__thd,  # type: ignore[arg-type]
            True,
            callable,
            *args,
            **kwargs,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
Support for buildsets in the database
"""

from __future__ import annotations

import json
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer

from buildbot.db import NULL
from buildbot.db import base
from buildbot.util import datetime2epoch
from buildbot.util import epoch2datetime
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    import datetime


class BsProps(dict):
    pass


class AlreadyCompleteError(RuntimeError):
    pass


@dataclass
class BuildSetModel:
    bsid: int
    external_idstring: str | None
    reason: str | None
    submitted_at: datetime.datetime
    complete: bool = False
    complete_at: datetime.datetime | None = None
    results: int | None = None
    parent_buildid: int | None = None
    parent_relationship: str | None = None
    rebuilt_buildid: int | None = None

    sourcestamps: list[int] = field(default_factory=list)

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'BuildsetsConnectorComponent '
                'getBuildset, getBuildsets, and getRecentBuildsets '
                'no longer return BuildSet as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


class BuildsetsConnectorComponent(base.DBConnectorComponent):
    @defer.inlineCallbacks
    def addBuildset(
        self,
        sourcestamps,
        reason,
        properties,
        builderids,
        waited_for,
        external_idstring=None,
        submitted_at=None,
        rebuilt_buildid=None,
        parent_buildid=None,
        parent_relationship=None,
        priority=0,
    ):
        # We've gotten this wrong a couple times.
        assert isinstance(waited_for, bool), f'waited_for should be boolean: {waited_for!r}'

        if submitted_at is not None:
            submitted_at = datetime2epoch(submitted_at)
        else:
            submitted_at = int(self.master.reactor.seconds())

        # convert to sourcestamp IDs first, as necessary
        def toSsid(sourcestamp):
            if isinstance(sourcestamp, int):
                return defer.succeed(sourcestamp)
            ssConnector = self.master.db.sourcestamps
            return ssConnector.findSourceStampId(**sourcestamp)

        sourcestamps = yield defer.DeferredList(
            [toSsid(ss) for ss in sourcestamps], fireOnOneErrback=True, consumeErrors=True
        )
        sourcestampids = [r[1] for r in sourcestamps]

        def thd(conn):
            buildsets_tbl = self.db.model.buildsets

            self.checkLength(buildsets_tbl.c.reason, reason)
            self.checkLength(buildsets_tbl.c.external_idstring, external_idstring)

            transaction = conn.begin()

            # insert the buildset itself
            r = conn.execute(
                buildsets_tbl.insert(),
                {
                    "submitted_at": submitted_at,
                    "reason": reason,
                    "rebuilt_buildid": rebuilt_buildid,
                    "complete": 0,
                    "complete_at": None,
                    "results": -1,
                    "external_idstring": external_idstring,
                    "parent_buildid": parent_buildid,
                    "parent_relationship": parent_relationship,
                },
            )
            bsid = r.inserted_primary_key[0]

            # add any properties
            if properties:
                bs_props_tbl = self.db.model.buildset_properties

                inserts = [
                    {"buildsetid": bsid, "property_name": k, "property_value": json.dumps([v, s])}
                    for k, (v, s) in properties.items()
                ]
                for i in inserts:
                    self.checkLength(bs_props_tbl.c.property_name, i['property_name'])

                conn.execute(bs_props_tbl.insert(), inserts)

            # add sourcestamp ids
            r = conn.execute(
                self.db.model.buildset_sourcestamps.insert(),
                [{"buildsetid": bsid, "sourcestampid": ssid} for ssid in sourcestampids],
            )

            # and finish with a build request for each builder.  Note that
            # sqlalchemy and the Python DBAPI do not provide a way to recover
            # inserted IDs from a multi-row insert, so this is done one row at
            # a time.
            brids = {}
            br_tbl = self.db.model.buildrequests
            ins = br_tbl.insert()
            for builderid in builderids:
                r = conn.execute(
                    ins,
                    {
                        "buildsetid": bsid,
                        "builderid": builderid,
                        "priority": priority,
                        "claimed_at": 0,
                        "claimed_by_name": None,
                        "claimed_by_incarnation": None,
                        "complete": 0,
                        "results": -1,
                        "submitted_at": submitted_at,
                        "complete_at": None,
                        "waited_for": 1 if waited_for else 0,
                    },
                )

                brids[builderid] = r.inserted_primary_key[0]

            transaction.commit()

            return (bsid, brids)

        bsid, brids = yield self.db.pool.do(thd)

        # Seed the buildset property cache.
        self.getBuildsetProperties.cache.put(bsid, BsProps(properties))

        return (bsid, brids)

    @defer.inlineCallbacks
    def completeBuildset(self, bsid, results, complete_at=None):
        if complete_at is not None:
            complete_at = datetime2epoch(complete_at)
        else:
            complete_at = int(self.master.reactor.seconds())

        def thd(conn):
            tbl = self.db.model.buildsets

            q = tbl.update().where(
                (tbl.c.id == bsid) & ((tbl.c.complete == NULL) | (tbl.c.complete != 1))
            )
            res = conn.execute(q.values(complete=1, results=results, complete_at=complete_at))
            conn.commit()

            if res.rowcount != 1:
                # happens when two buildrequests finish at the same time
                raise AlreadyCompleteError()

        yield self.db.pool.do(thd)

    def getBuildset(self, bsid) -> defer.Deferred[BuildSetModel | None]:
        def thd(conn) -> BuildSetModel | None:
            bs_tbl = self.db.model.buildsets
            q = bs_tbl.select().where(bs_tbl.c.id == bsid)
            res = conn.execute(q)
            row = res.fetchone()
            if not row:
                return None
            return self._thd_model_from_row(conn, row)

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def getBuildsets(self, complete=None, resultSpec=None):
        def thd(conn) -> list[BuildSetModel]:
            bs_tbl = self.db.model.buildsets
            q = bs_tbl.select()
            if complete is not None:
                if complete:
                    q = q.where(bs_tbl.c.complete != 0)
                else:
                    q = q.where((bs_tbl.c.complete == 0) | (bs_tbl.c.complete == NULL))
            if resultSpec is not None:
                return resultSpec.thd_execute(conn, q, lambda x: self._thd_model_from_row(conn, x))
            res = conn.execute(q)
            return [self._thd_model_from_row(conn, row) for row in res.fetchall()]

        res = yield self.db.pool.do(thd)
        return res

    def getRecentBuildsets(
        self,
        count: int | None = None,
        branch: str | None = None,
        repository: str | None = None,
        complete: bool | None = None,
    ) -> defer.Deferred[list[BuildSetModel]]:
        def thd(conn) -> list[BuildSetModel]:
            bs_tbl = self.db.model.buildsets
            ss_tbl = self.db.model.sourcestamps
            j = self.db.model.buildsets
            j = j.join(self.db.model.buildset_sourcestamps)
            j = j.join(self.db.model.sourcestamps)
            q = sa.select(bs_tbl).select_from(j).distinct()
            q = q.order_by(sa.desc(bs_tbl.c.submitted_at))
            q = q.limit(count)

            if complete is not None:
                if complete:
                    q = q.where(bs_tbl.c.complete != 0)
                else:
                    q = q.where((bs_tbl.c.complete == 0) | (bs_tbl.c.complete == NULL))
            if branch:
                q = q.where(ss_tbl.c.branch == branch)
            if repository:
                q = q.where(ss_tbl.c.repository == repository)
            res = conn.execute(q)
            return list(reversed([self._thd_model_from_row(conn, row) for row in res.fetchall()]))

        return self.db.pool.do(thd)

    @base.cached("BuildsetProperties")
    def getBuildsetProperties(self, bsid) -> defer.Deferred[BsProps]:
        def thd(conn) -> BsProps:
            bsp_tbl = self.db.model.buildset_properties
            q = sa.select(
                bsp_tbl.c.property_name,
                bsp_tbl.c.property_value,
            ).where(bsp_tbl.c.buildsetid == bsid)
            ret = []
            for row in conn.execute(q):
                try:
                    properties = json.loads(row.property_value)
                    ret.append((row.property_name, tuple(properties)))
                except ValueError:
                    pass
            return BsProps(ret)

        return self.db.pool.do(thd)

    def _thd_model_from_row(self, conn, row):
        # get sourcestamps
        tbl = self.db.model.buildset_sourcestamps
        sourcestamps = [
            r.sourcestampid
            for r in conn.execute(
                sa.select(tbl.c.sourcestampid).where(tbl.c.buildsetid == row.id)
            ).fetchall()
        ]

        return BuildSetModel(
            bsid=row.id,
            external_idstring=row.external_idstring,
            reason=row.reason,
            submitted_at=epoch2datetime(row.submitted_at),
            complete=bool(row.complete),
            complete_at=epoch2datetime(row.complete_at),
            results=row.results,
            parent_buildid=row.parent_buildid,
            parent_relationship=row.parent_relationship,
            rebuilt_buildid=row.rebuilt_buildid,
            sourcestamps=sourcestamps,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Callable

import sqlalchemy as sa

from buildbot.db import base
from buildbot.util.sautils import hash_columns
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from sqlalchemy.future.engine import Connection

    from buildbot.data.resultspec import ResultSpec


@dataclass
class CodebaseBranchModel:
    id: int
    codebaseid: int
    name: str
    commitid: int | None
    last_timestamp: int


class CodebaseBranchConnectorComponent(base.DBConnectorComponent):
    @async_to_deferred
    async def get_branch(self, id: int) -> CodebaseBranchModel | None:
        def thd(conn: Connection) -> CodebaseBranchModel | None:
            tbl = self.db.model.codebase_branches
            q = tbl.select().where(tbl.c.id == id)
            res = conn.execute(q)
            row = res.fetchone()
            rv = self._model_from_row(row) if row else None
            res.close()
            return rv

        return await self.db.pool.do(thd)

    @async_to_deferred
    async def get_branch_by_name(self, codebaseid: int, name: str) -> CodebaseBranchModel | None:
        def thd(conn: Connection) -> CodebaseBranchModel | None:
            tbl = self.db.model.codebase_branches
            q = tbl.select().where((tbl.c.codebaseid == codebaseid) & (tbl.c.name == name))
            res = conn.execute(q)
            row = res.fetchone()
            rv = self._model_from_row(row) if row else None
            res.close()
            return rv

        return await self.db.pool.do(thd)

    @async_to_deferred
    async def get_branches(
        self, codebaseid: int, result_spec: ResultSpec | None = None
    ) -> list[CodebaseBranchModel]:
        def thd(conn: Connection) -> list[CodebaseBranchModel]:
            tbl = self.db.model.codebase_branches
            q = tbl.select().where(tbl.c.codebaseid == codebaseid)

            if result_spec is not None:
                return result_spec.thd_execute(conn, q, self._model_from_row)

            res = conn.execute(q)
            rv = [self._model_from_row(row) for row in res.fetchall()]
            res.close()
            return rv

        return await self.db.pool.do(thd)

    @async_to_deferred
    async def update_branch(
        self,
        *,
        codebaseid: int,
        name: str,
        commitid: int | None = None,
        last_timestamp: int,
        _race_hook: Callable[[Connection], None] | None = None,
    ) -> None:
        tbl = self.db.model.codebase_branches
        self.checkLength(tbl.c.name, name)

        def thd(conn: Connection) -> None:
            try:
                self.db.upsert(
                    conn,
                    tbl,
                    where_values=(
                        (tbl.c.codebaseid, codebaseid),
                        (tbl.c.name_hash, hash_columns(name)),
                    ),
                    update_values=(
                        (tbl.c.commitid, commitid),
                        (tbl.c.name, name),
                        (tbl.c.last_timestamp, last_timestamp),
                    ),
                    _race_hook=_race_hook,
                )
                conn.commit()
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                # someone else beat us to the punch inserting this row;
                # let them win.
                conn.rollback()

        await self.db.pool.do_with_transaction(thd)

    def _model_from_row(self, row: sa.Row) -> CodebaseBranchModel:
        return CodebaseBranchModel(
            id=row.id,
            codebaseid=row.codebaseid,
            name=row.name,
            commitid=row.commitid,
            last_timestamp=row.last_timestamp,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test import fakedb
from buildbot.test.fake import fakemaster

if TYPE_CHECKING:
    from twisted.trial import unittest

    _ChangeSourceMixinBase = unittest.TestCase
else:
    _ChangeSourceMixinBase = object


class ChangeSourceMixin(_ChangeSourceMixinBase):
    """
    This class is used for testing change sources, and handles a few things:

     - starting and stopping a ChangeSource service
     - a fake master with a data API implementation
    """

    changesource = None
    started = False

    DUMMY_CHANGESOURCE_ID = 20
    OTHER_MASTER_ID = 93
    DEFAULT_NAME = "ChangeSource"

    @defer.inlineCallbacks
    def setUpChangeSource(self, want_real_reactor: bool = False):
        "Set up the mixin - returns a deferred."
        self.master = yield fakemaster.make_master(
            self, wantDb=True, wantData=True, wantRealReactor=want_real_reactor
        )

        self.master.db.insert_test_data([
            fakedb.Master(id=fakedb.FakeDBConnector.MASTER_ID),
            fakedb.Master(id=self.OTHER_MASTER_ID, active=1),
            fakedb.ChangeSource(id=self.DUMMY_CHANGESOURCE_ID, name=self.DEFAULT_NAME),
        ])

        @defer.inlineCallbacks
        def cleanup():
            if not self.started:
                return
            if self.changesource.running:
                yield self.changesource.stopService()
            yield self.changesource.disownServiceParent()

        self.addCleanup(cleanup)

    @defer.inlineCallbacks
    def attachChangeSource(self, cs):
        self.changesource = cs
        yield self.changesource.setServiceParent(self.master)
        yield self.changesource.configureService()
        return cs

    def startChangeSource(self):
        "start the change source as a service"
        self.started = True
        return self.changesource.startService()

    @defer.inlineCallbacks
    def stopChangeSource(self):
        "stop the change source again; returns a deferred"
        yield self.changesource.stopService()

        self.started = False

    @defer.inlineCallbacks
    def setChangeSourceToMaster(self, otherMaster):
        ret = yield self.master.data.updates.trySetChangeSourceMaster(
            self.DUMMY_CHANGESOURCE_ID, otherMaster
        )
        return ret



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from twisted.internet import defer

from buildbot.util import ComparableMixin
from buildbot.util import subscription
from buildbot.util.eventual import eventually


class Listener:
    pass


class UpdateRegistrationListener(Listener):
    def __init__(self):
        super().__init__()
        # username : (password, portstr, manager registration)
        self._registrations = {}

    @defer.inlineCallbacks
    def updateRegistration(self, username, password, portStr):
        # NOTE: this method is only present on the PB and MsgPack protocols; others do not
        # use registrations
        if username in self._registrations:
            currentPassword, currentPortStr, currentReg = self._registrations[username]
        else:
            currentPassword, currentPortStr, currentReg = None, None, None

        iseq = ComparableMixin.isEquivalent(
            currentPassword, password
        ) and ComparableMixin.isEquivalent(currentPortStr, portStr)
        if iseq:
            return currentReg
        if currentReg:
            yield currentReg.unregister()
            del self._registrations[username]

        if portStr is not None and password:
            reg = yield self.get_manager().register(
                portStr, username, password, self._create_connection
            )
            self._registrations[username] = (password, portStr, reg)
            return reg
        return currentReg

    @defer.inlineCallbacks
    def _create_connection(self, mind, workerName):
        self.before_connection_setup(mind, workerName)
        worker = self.master.workers.getWorkerByName(workerName)
        conn = self.ConnectionClass(self.master, worker, mind)

        # inform the manager, logging any problems in the deferred
        accepted = yield self.master.workers.newConnection(conn, workerName)

        # return the Connection as the perspective
        if accepted:
            return conn
        else:
            # TODO: return something more useful
            raise RuntimeError("rejecting duplicate worker")


class Connection:
    proxies: dict[type, type] = {}

    def __init__(self, name):
        self._disconnectSubs = subscription.SubscriptionPoint(f"disconnections from {name}")

    # This method replace all Impl args by their Proxy protocol implementation
    def createArgsProxies(self, args):
        newargs = {}
        for k, v in args.items():
            for implclass, proxyclass in self.proxies.items():
                if isinstance(v, implclass):
                    v = proxyclass(v)
            newargs[k] = v
        return newargs

    def get_peer(self):
        raise NotImplementedError

    # disconnection handling

    def wait_shutdown_started(self):
        d = defer.Deferred()
        self.notifyOnDisconnect(lambda: eventually(d.callback, None))
        return d

    def waitShutdown(self):
        return self._disconnectSubs.waitForDeliveriesToFinish()

    def notifyOnDisconnect(self, cb):
        return self._disconnectSubs.subscribe(cb)

    def notifyDisconnected(self):
        self._disconnectSubs.deliver()

    def loseConnection(self):
        raise NotImplementedError

    # methods to send messages to the worker

    def remotePrint(self, message):
        raise NotImplementedError

    def remoteGetWorkerInfo(self):
        raise NotImplementedError

    def remoteSetBuilderList(self, builders):
        raise NotImplementedError

    def remoteStartCommand(self, remoteCommand, builderName, commandId, commandName, args):
        raise NotImplementedError

    def remoteShutdown(self):
        raise NotImplementedError

    def remoteStartBuild(self, builderName):
        raise NotImplementedError

    def remoteInterruptCommand(self, builderName, commandId, why):
        raise NotImplementedError


# RemoteCommand base implementation and base proxy
class RemoteCommandImpl:
    def remote_update(self, updates):
        raise NotImplementedError

    def remote_complete(self, failure=None):
        raise NotImplementedError


# FileWriter base implementation
class FileWriterImpl:
    def remote_write(self, data):
        raise NotImplementedError

    def remote_utime(self, accessed_modified):
        raise NotImplementedError

    def remote_unpack(self):
        raise NotImplementedError

    def remote_close(self):
        raise NotImplementedError


# FileReader base implementation
class FileReaderImpl:
    def remote_read(self, maxLength):
        raise NotImplementedError

    def remote_close(self):
        raise NotImplementedError
